# Code Documentation for Network Partitioning and Distributed Control Modules

## Overview

This repository contains a set of Python modules designed for network partitioning, ADMM-based distributed model predictive control (DMPC-ADMM), and performance evaluation of the resulting strategies, including metrics comparison and graphical representations. The framework supports both algorithmic and optimization-based partitioning, enabling efficient non-centralized control of large-scale dynamical systems.

The repository implements the methodologies proposed and explained in the paper [1]. The techniques are based on the core concepts of Fundamental System Units (FSUs), the smallest indivisible sub-systems, and their aggregation into Composite System Units (CSUs), which serve as the basis for partitioning.

## Core Features

* Selection FSUs: implements an algorithm for extracting FSUs from a given network based on topological and dynamics information.

* Partitioning by aggregation of CSUs: implements algorithmic and optimization-based techniques to aggregate FSUs into CSUs, forming control partitions.

* DMPC-ADMM: implements ADMM-based DMPC for both networks of linear and hybrid systems. The DMPC implementation is based on a prior assignment of a partition. Consensus structures, as well as the definition of local optimization problems, are all automated once the network topology and partition are specified.

* Performance Evaluation: comparative analysis of partitioning strategies using computation time [s], computation cost [core-seconds], and cost function values.

* Visualization: provides graphical representation of network topologies and partitions using circular graph representations, control performance, and system evolution.

## Selection of Fundamental System Units (FSUs)

The partitioning process begins with the selection of FSUs. These are the smallest subsystems that can operate independently within the constraints of the network dynamics. FSUs cannot be split further without violating the non-autonomicity assumption. They serve as the building blocks for defining CSUs in the next phase of partitioning.

This selection process ensures that all control-relevant interactions are preserved while minimizing interconnections between partitions, for a given granularity level.

## Partitioning: Aggregating FSUs into CSUs

Once FSUs are selected, they are aggregated into composite system units (CSUs) to create a structured control partition. This process is guided by two distinct methodologies:

### 1. Algorithmic Partitioning

Algorithmic partitioning follows a greedy approach to iteratively form CSUs while optimizing intra-CSU connectivity, minimizing inter-CSU interactions, and accounting for the desired granularity of the partitioning.

### 2. Optimization-Based Partitioning

Optimization-based partitioning formulates the problem as a integer quadratic programming (IQP) allowing for an optimal CSU configuration according to the quadratic partiton index.

This method provides globally optimal partitions but at a significantly higher computational cost compared to heuristic approaches.

## Modules and Functions

### 1. A00_Support_Functions.py
This module provides general-purpose functions used throughout the framework.
* `simulation_parameters(NUMBER_ATOMIC_AGENTS: int) -> dict`
  -- Initializes simulation parameters, including prediction horizon, control constraints, and weighting matrices.

### 2. A00_Random_Network_Generation.py
Generates randomized network topologies to simulate different networks, returns the associated system dynamics. 
* `random_network_generator(number_of_states: int, number_of_inputs: int, number_of_elements_per_column: int) -> tuple`
  -- Generates random adjacency matrices and state-space representations.

### 3. A01_Network_Generator.py
Constructs structured networks using weighted adjacency matrices and provides visualization tools for graph representation.

### 4. A02_Weighting_Function.py
Provides functions to evaluate partitioning quality by analyzing intra- and inter-CSU interactions.
* `weighting_function(Network_Dict: dict, List_Atomic_Agents: list, Partition: list, ALPHA: float) -> float`
   -- Computes partition index based on on a given partition, uses the metrics defined in [1].
  
### 5. A03_Algorithmic_Partitioning.py
Implements a greedy partitioning algorithm to iteratively merge FSUs into CSUs based on topology and weighting functions according to the nonlinear partition index [1].

### 6. A03_Algorithmic_Partitioning_Mixed.py
Enhances the greedy partitioning algorithm by refining CSU assignments through local FSUs exchanges.

### 7. A09_Network_Partitioning_Quadratic.py
Solves an IQP problem to find an optimal network partitioning according to the quadratic partition index in [1].

### 8. A13_Manual_Partition_Definition.py
Allows users to define partitions manually, for comparative and benchmarking purposes.

### 9. A04_DMPC_ADMM_Linear.py
Implements DMPC using ADMM for linear systems, ensuring distributed optimization across CSUs.

### 10. A05_DMPC_ADMM_Hybrid.py
Extends DMPC to hybrid (PWA) systems, incorporating mixed-integer constraints for handling logic-based system transitions.

### 11. A16_DMPC_Random_50_Results_Comparison.py
Analyzes and compares performance metrics across different partitioning methods for DMPC simulations on a Random_50 network.

### 12. A10_Plot_Graph_Partitions.py
Provides graph-based visualization for analyzing CSU structures and control performance outcomes.

## Dependencies
* `numpy` – Matrix computations
* `pickle` – Data serialization
* `networkx` – Graph structures and visualization
* `matplotlib` – Graph-based visualization
* `pandas` – Structured data handling
* `gurobipy` – Optimization solvers

## Bibliography
[1] A. Riccardi, L. Laurenti, and B. De Schutter, “A general partitioning strategy for non-centralized control,” 2025, to appear.



